/*
 *  aRts ouput plugin for xmms
 *
 *  Copyright (C) 2000,2003  Haavard Kvaalen <havardk@xmms.org>
 *
 *  Licenced under GNU GPL version 2.
 */

#include "arts.h"
#include <errno.h>
#include <signal.h>
#include <pthread.h>

static arts_stream_t handle;

static gboolean going, paused;
static int arts_buffer, latency, psize;
static guint64 written = 0;
static struct params_info input_params, output_params;

static int (*arts_convert_func)(void **data, int length);
struct arts_config artsxmms_cfg;

static pthread_mutex_t artsm = PTHREAD_MUTEX_INITIALIZER;


struct {
	int left, right;
} volume = {100, 100};

void artsxmms_init(void)
{
	ConfigFile *cfgfile;

	memset(&artsxmms_cfg, 0, sizeof (artsxmms_cfg));

	artsxmms_cfg.buffer_size = 1000;
	
	cfgfile = xmms_cfg_open_default_file();
	xmms_cfg_read_int(cfgfile, "arts", "buffer_size",
			  &artsxmms_cfg.buffer_size);
	xmms_cfg_free(cfgfile);
}

static void artsxmms_set_params(struct params_info *params, AFormat fmt, int rate, int nch)
{
	params->format = fmt;
	params->frequency = rate;
	params->channels = nch;

	params->bps = rate * nch;
	params->resolution = 8;
	if (!(fmt == FMT_U8 || fmt == FMT_S8))
	{
		params->bps *= 2;
		params->resolution = 16;
	}
}

int artsxmms_get_written_time(void)
{
	if (!going)
		return 0;

	return (written * 1000) / output_params.bps;
}

int artsxmms_get_output_time(void) 
{
	int time, space;
	guint64 w;
	
	if (!going)
		return 0;

	pthread_mutex_lock(&artsm);
	w = written;
	space = arts_stream_get(handle, ARTS_P_BUFFER_SPACE);
	if (arts_buffer - space > w)
		w = 0;
	else
		w -= arts_buffer - space;
	time =  (w * 1000) / output_params.bps;

	if (!paused)
	{
		if (time < latency)
			time = 0;
		else
			time -= latency;
	}

	pthread_mutex_unlock(&artsm);
	return time;
}

int artsxmms_playing(void)
{
	if (!going)
		return FALSE;
	
	if (!paused)
	{
		int t;
		pthread_mutex_lock(&artsm);
		t = arts_stream_get(handle, ARTS_P_BUFFER_SPACE);
		pthread_mutex_unlock(&artsm);
		return t < arts_buffer - psize;
	}

	return TRUE;
}

int artsxmms_free(void)
{
	int space;

	if (paused || !going)
		return 0;

	pthread_mutex_lock(&artsm);
	space = arts_stream_get(handle, ARTS_P_BUFFER_SPACE); 
	pthread_mutex_unlock(&artsm);

	return space;
}

static void artsxmms_open_stream(struct params_info *params)
{
	int buffersize = artsxmms_cfg.buffer_size / 1000 * params->bps;
	handle = arts_play_stream(params->frequency, params->resolution,
				  params->channels, "XMMS");
	arts_buffer = arts_stream_set(handle, ARTS_P_BUFFER_SIZE, buffersize);
	latency = arts_stream_get(handle, ARTS_P_SERVER_LATENCY);
	psize = arts_stream_get(handle, ARTS_P_PACKET_SIZE);

	arts_convert_func = arts_get_convert_func(params->format);
}

static void volume_adjust(void* data, int length)
{
	int i;

	if ((volume.left == 100 && volume.right == 100) ||
	    (output_params.resolution == 8 &&
	     (volume.left == 100 || volume.right == 100)))
		return;

	if (output_params.resolution == 16)
	{
		gint16 *ptr = data;
		if (output_params.channels == 2)
			for (i = 0; i < length; i += 4)
			{
				*ptr = GINT16_TO_LE(GINT16_FROM_LE(*ptr) *
						    volume.left / 100);
				ptr++;
				*ptr = GINT16_TO_LE(GINT16_FROM_LE(*ptr) *
						    volume.right / 100);
				ptr++;
			}
		else
		{
			int vol = MAX(volume.left, volume.right);
			for (i = 0; i < length; i += 2, ptr++)
			{
				*ptr = GINT16_TO_LE(GINT16_FROM_LE(*ptr) *
						    vol / 100);
			}
		}
	}
	else
	{
		guint8 *ptr = data;
		if (output_params.channels == 2)
			for (i = 0; i < length; i += 2)
			{
				*ptr = *ptr * volume.left / 100;
				ptr++;
				*ptr = *ptr * volume.right / 100;
				ptr++;
			}
		else
		{
			int vol = MAX(volume.left, volume.right);
			for (i = 0; i < length; i++, ptr++)
			{
				*ptr = *ptr * vol / 100;
			}
		}
	}
}

void artsxmms_write(gpointer ptr, int length)
{
	AFormat new_format;
	int new_frequency, new_channels;
	EffectPlugin *ep;
	int ret;
	
	new_format = output_params.format;
	new_frequency = output_params.frequency;
	new_channels = output_params.channels;

	ep = get_current_effect_plugin();
	if (effects_enabled() && ep && ep->query_format)
		ep->query_format(&new_format, &new_frequency, &new_channels);
	
	pthread_mutex_lock(&artsm);

	if (new_format != output_params.format ||
	    new_frequency != output_params.frequency ||
	    new_channels != output_params.channels)
	{
		//FIXME written
		/*
		 * The effect plugins can change the format of the stream.
		 */
		
		arts_close_stream(handle);
		artsxmms_set_params(&output_params, new_format,
				    new_frequency, new_channels);
		artsxmms_open_stream(&output_params);
	}
	if (effects_enabled() && ep && ep->mod_samples)
		length = ep->mod_samples(&ptr, length, input_params.format,
					 input_params.frequency,
					 input_params.channels);

	if (arts_convert_func)
		arts_convert_func(ptr, length);

	volume_adjust(ptr, length);

	ret = arts_write(handle, ptr, length);
	if (ret < 0)
	{
		/* FIXME: handle this better? */
		g_message("artsxmms_write(): write error: %s",
			  arts_error_text(ret));
	}
	else
		written += ret;
	pthread_mutex_unlock(&artsm);
}

void artsxmms_close(void)
{
	pthread_mutex_lock(&artsm);
	going = 0;
	arts_close_stream(handle);
	arts_free();
	pthread_mutex_unlock(&artsm);
}

void artsxmms_flush(int time)
{
	/*
	 * Argh, no way to flush the stream from the C api.
	 */
	written = (time / 10) * (output_params.bps / 100);
}

void artsxmms_pause(short p)
{
	paused = p;
}

int artsxmms_open(AFormat fmt, int rate, int nch)
{
	int err;

	pthread_mutex_lock(&artsm);
	/* Stop the arts initialization from complaining */
	signal(SIGPIPE, SIG_DFL);
	if ((err = arts_init()) != 0)
	{
		g_message("artsxmms_open(): Unable to initialize aRts: %s",
			  arts_error_text(err));
		pthread_mutex_unlock(&artsm);
		return 0;
	}

	artsxmms_set_params(&input_params, fmt, rate, nch);
	artsxmms_set_params(&output_params, fmt, rate, nch);

	written = 0;
	paused = 0;

	artsxmms_open_stream(&output_params);

	going = 1;
	pthread_mutex_unlock(&artsm);
	return 1;
}

void artsxmms_get_volume(int *l, int *r)
{
	*l = volume.left;
	*r = volume.right;
}

void artsxmms_set_volume(int l, int r)
{
	volume.left = l;
	volume.right = r;
}
